#!/bin/sh
# the next line restarts using wish \
exec wish "$0" "$@"

# main.tcl --
#
#       Interactive simulation of an M/M/1 queue.
#
#       This is the start-up script that implements the Tcl/Tk code for the
#       GUI.
#
# Copyright (c) 2008 Antonio Bonifati
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
# USA.
#
# Changelog:
#
# 12/14/2008: Added log capabilities.

set auto_path [linsert $auto_path 0 .]
package require mm1Queue

#
# Button actions
#

proc quit {} {
    destroy .
}

proc reset {} {
    global state

    .buttons.reset configure -state disabled
    set state notStarted
    mm1Queue::pause
    .buttons.runPause configure -text Run
    mm1Queue::reset

    after cancel keepUpdateInt
    deleteInt

    after cancel keepUpdateMean
    deleteMean

    after cancel keepUpdateTime
    updateTime

    after cancel keepUpdateLog
}

proc runPause {} {
    global state

    .buttons.reset configure -state normal
    if {$state!="running"} {
        set state "running"
        mm1Queue::run
        .buttons.runPause configure -text Pause
        keepUpdateInt
        keepUpdateMean
        keepUpdateTime
        keepUpdateLog
    } else {
        set state "stopped"
        mm1Queue::pause
        .buttons.runPause configure -text Run
        after cancel keepUpdateInt
        after cancel keepUpdateMean
        after cancel keepUpdateTime
        after cancel keepUpdateLog
    }
}

proc oneStep {} {
    global state

    .buttons.reset configure -state normal
    if {$state!="running"} {
        mm1Queue::oneStep
        updateInt
        updateMean
        updateTime
        updateLog
    } else {
        set state "stopped"
        mm1Queue::pause
        .buttons.runPause configure -text Run
        mm1Queue::oneStep
        after cancel keepUpdateInt
        updateInt

        after cancel keepUpdateMean
        updateMean

        after cancel keepUpdateTime
        updateTime

        after cancel keepUpdateLog
        updateLog
    }
}

#
# Create all windows, and pack them
#

#
# Model parameters
#

proc toggleModelParam {} {
    global paramMapped

    if {$paramMapped} {
        set paramMapped no

        destroy .param
    } else {
        set paramMapped yes

        modelParam
    }
}

proc modelParam {} {
    toplevel .param
    wm group .param .
    wm title .param "M/M/1 model parameters"
    wm resizable .param no no
    wm protocol .param WM_DELETE_WINDOW toggleModelParam
    focus .param

    frame .param.frame

    label .param.frame.arrRateLabel -text "Arrival Rate (\u03bb)" -anchor e
    spinbox .param.frame.arrRateEntry -width 10 -from 0 -to 1000
    grid .param.frame.arrRateLabel .param.frame.arrRateEntry -sticky ew -pady 2 -padx 1

    label .param.frame.servRateLabel -text "Service Rate (\u00b5)" -anchor e
    spinbox .param.frame.servRateEntry -width 10 -from 0 -to 1000
    grid .param.frame.servRateLabel .param.frame.servRateEntry -sticky ew -pady 2 -padx 1

    frame .param.buttons
    button .param.buttons.close -text Close -default active -command toggleModelParam
    button .param.buttons.reset -text Reset -default normal -command {
        .param.frame.arrRateEntry set [mm1Queue::mm1set arrRate]
        .param.frame.servRateEntry set [mm1Queue::mm1set servRate]
    }
    button .param.buttons.change -text Change -default normal -command {
        set oldArrRate [mm1Queue::mm1set arrRate]
        set oldServRate [mm1Queue::mm1set servRate]
        if {[catch {mm1Queue::mm1set arrRate [.param.frame.arrRateEntry get]} errmsg] \
                || [catch {mm1Queue::mm1set servRate [.param.frame.servRateEntry get]} errmsg]} {
            tk_messageBox -title mm1 -icon error -message $errmsg -parent .param
        } elseif {[mm1Queue::isStable] || [tk_messageBox -title mm1 -icon question -type yesno -parent .param \
                -message "Your system is not stable! For a stable system the Arrival rate must be smaller than the Service rate.\n\nProceed anyway?"]} {
            updateLimit
            rates
        } else {
            #
            # Restore parameters.
            #

            mm1Queue::mm1set arrRate $oldArrRate
            mm1Queue::mm1set servRate $oldServRate
        }
    }
    pack .param.buttons.close .param.buttons.reset .param.buttons.change -side left -padx 5

    bind .param <Return> {.param.buttons.close flash; .param.buttons.close invoke}
    bind .param <Escape> {.param.buttons.close flash; .param.buttons.close invoke}

    .param.buttons.reset invoke

    pack .param.frame -side top -fill y -anchor center
    pack .param.buttons -side bottom -pady 2 -anchor center
}

proc updateLog {} {
    global logEnabled fileChannel logFile varsToLog meanMapped

    if {$logEnabled && $varsToLog !=  ""} {
        if {$fileChannel == ""} {
            # default error handling seems to be appropriate, no catch needed
            set fileChannel [open $logFile w]

            # print headers
            puts $fileChannel [join $varsToLog ,]
        }

        if {! $meanMapped} {
            mm1Queue::evaluate
        }

        set logLine {}
        foreach varName $varsToLog {
            lappend logLine [mm1Queue::mm1set $varName]
        }
        puts $fileChannel [join $logLine ,]
    }
}

proc updateInt {} {
    global intMapped intVar

    if {$intMapped} {
        foreach varName $intVar {
            .int.frame.${varName}Value configure -text [format "%g" [mm1Queue::mm1set $varName]]
        }
    }
}

proc keepUpdateInt {} {
    global intMapped updateRate

    if {$intMapped} {
        updateInt
        after $updateRate keepUpdateInt
    }
}

proc keepUpdateLog {} {
    global logEnabled updateRate

    if {$logEnabled} {
        updateLog
        after $updateRate keepUpdateLog
    }
}

proc deleteInt {} {
    global intMapped intVar

    if {$intMapped} {
        foreach varName $intVar {
            .int.frame.${varName}Value configure -text "-"
        }
    }
}

#
# Max, min and integers
#

proc toggleIntResults {} {
    global intMapped

    if {$intMapped} {
        set intMapped no

        destroy .int
    } else {
        set intMapped yes

        intResults
    }
}

proc intResults {} {
    global intVar intVarLabel state updateRate

    toplevel .int
    wm group .int .
    wm title .int "M/M/1 max, min and integers"
    wm resizable .int no no
    wm protocol .int WM_DELETE_WINDOW toggleIntResults
    focus .int

    label .int.label -text "Values updated every ${updateRate}ms" -font "Helvetica -12 italic"

    frame .int.frame
    foreach varName $intVar label $intVarLabel {
        label .int.frame.${varName}Label -text $label: -anchor nw -pady 0
        label .int.frame.${varName}Value -anchor ne -width 10 -pady 0
        grid .int.frame.${varName}Label .int.frame.${varName}Value -sticky ew -pady 0
        grid configure .int.frame.${varName}Label -padx 1m
    }
    grid columnconfigure .int.frame 1 -weight 1

    button .int.close -text Close -default active -command toggleIntResults

    bind .int <Return> {.int.close flash; .int.close invoke}
    bind .int <Escape> {.int.close flash; .int.close invoke}

    pack .int.label -side top
    pack .int.close -side bottom -pady 2m
    pack .int.frame -expand 1 -fill both -padx 2m -pady 1m

    if {$state=="running"} {
        keepUpdateInt
    } elseif {$state=="stopped"} {
        updateInt
    } else {
        deleteInt
    }
}

#
# Average values and probabilities
#

proc toggleMeanResults {} {
    global meanMapped

    if {$meanMapped} {
        set meanMapped no

        destroy .mean
    } else {
        set meanMapped yes

        meanResults
    }
}

proc meanResults {} {
    global meanVar limitVar limitVarLabel state updateRate

    toplevel .mean
    wm group .mean .
    wm title .mean "M/M/1 mean results"
    wm resizable .mean no no
    wm protocol .mean WM_DELETE_WINDOW toggleMeanResults
    focus .mean

    label .mean.label -text "Values updated every ${updateRate}ms" -font "Helvetica -12 italic"

    frame .mean.frame
    label .mean.frame.simulated -text simulated -anchor ne -width 10 -pady 0
    label .mean.frame.limit -text limit -anchor ne -width 10 -pady 0
    grid x .mean.frame.simulated .mean.frame.limit
    foreach varName $meanVar limitVarName $limitVar label $limitVarLabel {
        label .mean.frame.${varName}Label -text $label: -anchor nw -pady 0
        label .mean.frame.${varName}Value -anchor ne -width 10 -pady 0
        label .mean.frame.$limitVarName -anchor ne -width 10 -pady 0
        grid .mean.frame.${varName}Label .mean.frame.${varName}Value .mean.frame.$limitVarName -sticky ew -pady 0
        grid configure .mean.frame.${varName}Label -padx 1m
    }
    grid columnconfigure .mean.frame 1 -weight 1

    button .mean.close -text Close -default active -command toggleMeanResults

    bind .mean <Return> {.mean.close flash; .mean.close invoke}
    bind .mean <Escape> {.mean.close flash; .mean.close invoke}

    pack .mean.label -side top
    pack .mean.close -side bottom -pady 2m
    pack .mean.frame -expand 1 -fill both -padx 2m -pady 1m

    updateLimit

    if {$state=="running"} {
        keepUpdateMean
    } elseif {$state=="stopped"} {
        updateMean
    } else {
        deleteMean
    }
}

proc updateMean {} {
    global meanMapped meanVar

    if {$meanMapped} {
        mm1Queue::evaluate
        foreach varName $meanVar {
            if {[set value [mm1Queue::mm1set $varName]] != ""} {
                .mean.frame.${varName}Value configure -text [format "%g" $value]
            }
        }
    }
}

proc updateLimit {} {
    global meanMapped limitVar

    if {$meanMapped} {
        if {[mm1Queue::isStable]} {
            foreach varName $limitVar {
                .mean.frame.$varName configure -text [format "%g" [mm1Queue::mm1set $varName]]
            }
        } else {
            foreach varName $limitVar varValue {1 0 "+\u221e" "+\u221e" "+\u221e" "+\u221e"} {
                .mean.frame.$varName configure -text $varValue
            }
        }
    }
}

proc keepUpdateMean {} {
    global meanMapped updateRate

    if {$meanMapped} {
        updateMean
        after $updateRate keepUpdateMean
    }
}

proc deleteMean {} {
    global meanMapped meanVar

    if {$meanMapped} {
        foreach varName $meanVar {
            .mean.frame.${varName}Value configure -text "-"
        }
    }
}

proc toggleLogOptions {} {
    global logMapped

    if {$logMapped} {
        set logMapped no

        destroy .log
    } else {
        set logMapped yes

        logOptions
    }
}

proc setLogFile {} {
    global logFile fileChannel varsToLog

    set types {
        {{Comma-separated values} csv TEXT}
        {{All Files}              *       }
    }
    set newFile [tk_getSaveFile -defaultextension csv -filetypes $types \
        -initialfile mm1Queue.csv -title "Log Result to File" -parent .log]

    if {$newFile != ""} {
        set logFile $newFile

        # ensure an eventually previous log file is properly closed
        if {$fileChannel != ""} {
            close $fileChannel
        }

        if {$varsToLog != ""} {
            .log.enableButton configure -state normal
        }
    }
}

proc updateVarsToLog {} {
    global varsToLog intVar meanVar logFile
    
    set varsToLog [list]
    set vars [concat {time} $intVar $meanVar]
    foreach var [.log.frame.varList curselection] {
        lappend varsToLog [lindex $vars $var]
    }

    .log.enableButton configure -state \
        [expr {$varsToLog != "" && $logFile != "" ? "normal" : "disabled"}]
}

proc setVarListSelection {} {
    global varsToLog intVar meanVar

    set vars [concat {time} $intVar $meanVar]
    foreach var $varsToLog {
        .log.frame.varList selection set [lsearch $vars $var]
    }
}

proc logOptions {} {
    global intVarLabel limitVarLabel

    toplevel .log
    wm group .log .
    wm title .log "M/M/1 log options"
    wm resizable .log no no
    wm protocol .log WM_DELETE_WINDOW toggleLogOptions
    focus .log

    label .log.fileLabel -text "Log File" -anchor e
    entry .log.fileEntry -textvariable logFile -state readonly -width 20
    button .log.fileButton -text "browse" -command setLogFile
    grid .log.fileLabel .log.fileEntry .log.fileButton -sticky ew -pady 2 -padx 1

    label .log.enableLabel -text "Log Enable" -anchor e
    checkbutton .log.enableButton -onvalue yes -offvalue no \
        -variable logEnabled -state disabled -command {
            if {$logEnabled} {
                keepUpdateLog
            }
        }
    grid .log.enableLabel .log.enableButton - -sticky ew -pady 2 -padx 1

    label .log.varLabel -text "Variables\nto Log"
    frame .log.frame
    listbox .log.frame.varList -selectmode multiple -exportselection no \
        -setgrid yes -yscrollcommand {.log.frame.varScroll set}
    # Better way, but unsupported by older Tcl versions:
    #.log.frame.varList insert end \
        {*}[concat {"simulated time"} $intVarLabel $limitVarLabel]
    eval .log.frame.varList insert end \
        [concat {"simulated time"} $intVarLabel $limitVarLabel]
    bind .log.frame.varList <<ListboxSelect>> updateVarsToLog
    setVarListSelection
    scrollbar .log.frame.varScroll -command {.log.frame.varList yview}
    pack .log.frame.varScroll -side right -fill y
    pack .log.frame.varList -expand yes -fill both
    grid .log.varLabel .log.frame - -sticky ew -pady 2 -padx 1
}

proc guy {c x y} {
    # head
    $c create oval $x $y [expr {$x + 4}] [expr {$y + 4}]
    # body
    $c create line [expr {$x + 2}] [expr {$y + 4}] [expr {$x + 2}] [expr {$y + 12}]
    # legs
    $c create line [expr {$x + 2}] [expr {$y + 12}] [expr {$x - 2}] [expr {$y + 19}]
    $c create line [expr {$x + 2}] [expr {$y + 12}] [expr {$x + 6}] [expr {$y + 19}]
    # arms
    $c create line [expr {$x + 2}] [expr {$y + 7}] [expr {$x - 2}] [expr {$y + 9}]
    $c create line [expr {$x + 2}] [expr {$y + 7}] [expr {$x + 7}] [expr {$y + 9}]
}

proc row {c x y n} {
    for {set i 0} {$i < $n} {incr i} {
        $c create rectangle $x $y [expr {$x+20}] [expr {$y+25}]
        incr x 20
    }
}   

proc rates {} {
    global mu lambda

    .draw.queue itemconfigure $lambda -text "\u03bb=[format "%g" [mm1Queue::mm1set arrRate]]"
    .draw.queue itemconfigure $mu -text "\u00b5=[format "%g" [mm1Queue::mm1set servRate]]"
}

proc updateTime {} {
    global time

    .draw.queue itemconfigure $time -text "time=[format "%g" [mm1Queue::mm1set time]]"
}

proc keepUpdateTime {} {
    global updateRate

    updateTime
    after $updateRate keepUpdateTime
}

proc createwindows {} {
    global paramMapped intMapped meanMapped mu lambda time

    #
    # Set various options
    #

    wm title . "M/M/1 queue simulator"
    wm resizable . no no
    option add *Label.font "Helvetica -12"

    #
    # Buttons
    #

    frame .buttons
    button .buttons.quit -text Quit -default normal -command quit
    button .buttons.reset -text Reset -default normal -state disabled -command reset
    button .buttons.runPause -text Run -width 6 -default active -command runPause
    button .buttons.oneStep -text "One step" -default normal -command oneStep
    menubutton .buttons.subWin -text "Subwindows" -indicator on -relief raised \
        -menu .buttons.subWin.menu -direction flush -highlightthickness 5
    menu .buttons.subWin.menu -tearoff no
    .buttons.subWin.menu add command -label "model parameters" -command toggleModelParam
    .buttons.subWin.menu add command -label "max, min and integers" -command toggleIntResults
    .buttons.subWin.menu add command -label "mean results" -command toggleMeanResults
    .buttons.subWin.menu add command -label "log options" -command toggleLogOptions
    pack .buttons.quit .buttons.reset .buttons.runPause .buttons.oneStep \
        .buttons.subWin -side left -expand yes -pady 5

    bind . <Return> {.buttons.runPause flash; .buttons.runPause invoke}
    bind . <Escape> {.buttons.quit flash; .buttons.quit invoke}

    #
    # Queue draw
    #
    frame .draw
    canvas .draw.queue -width 330 -height 70
    guy .draw.queue 20 10
    guy .draw.queue 35 10
    .draw.queue create line 50 20 70 20 -arrow last
    row .draw.queue 75 7 7
    guy .draw.queue 163 10
    guy .draw.queue 183 10
    guy .draw.queue 203 10
    .draw.queue create line 215 19 235 19 -arrow last
    guy .draw.queue 245 10
    .draw.queue create oval 235 7 260 32
    .draw.queue create line 260 19 285 19 -arrow last
    guy .draw.queue 295 10
    guy .draw.queue 315 10
    set lambda [.draw.queue create text 30 45]
    set mu [.draw.queue create text 280 45]
    set time [.draw.queue create text 150 45]
    rates
    updateTime
    pack .draw.queue

    #
    # Main window
    #

    pack .buttons -side top -fill both
    pack .draw -side top -fill both -expand yes -padx 1m -pady 1m
    if {$paramMapped} {
        modelParam
    }
    if {$intMapped} {
        intResults
    }
    if {$meanMapped} {
        meanResults
    }
}

##############################################################################
# Main program

#
# Initial state of simulation.
#

set updateRate 500
set paramMapped no
set intMapped no
set meanMapped no
set logMapped no
set state notStarted
set intVar {qlen numArr noWait minInt maxInt numServ minServ maxServ maxQlen maxWait maxSystem}
set intVarLabel {{queue length} {number of arrivals} \
   {number of not waiting arrivals} {minimum arrival interval} \
   {maximum arrival interval} {number of services} \
   {minimum service duration} {maximum service duration} \
   {maximum queue length} {maximum waiting time} \
   {maximum time in system}}
set meanVar {roSim p0Sim lsSim lqSim wsSim wqSim}
set limitVar {roLimit p0Limit lsLimit lqLimit wsLimit wqLimit}
set limitVarLabel {{server utilization} {idle probability} \
    {customers in system} {customers in queue} \
    {time in system} {time in queue}}
set logEnabled no
set logFile {}
set fileChannel {}
# Use variable names from intVar and meanVar lists (plus the
# global time) here to set the default variables to log
set varsToLog {}

#
# Create all windows
#

createwindows

#
# Now, wait for user actions...
#

# End of file
# vim: set ts=4 sw=4 et:
