# mm1Queue.tcl --
#
#       Interactive simulation of an M/M/1 queue.
#
#       This is the primary file in the mm1Queue library and implements
#       the Tcl code for both exported and internal commands.
#
# Copyright (c) 2005 Antonio Bonifati
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
# USA.
#
# $Id$

package provide mm1Queue 0.1

namespace eval mm1Queue {
    namespace export expRand mm1set isStable reset run pause oneStep evaluate

    #
    # Model parameters.
    #

    # Mean rate of customers arrivals or "lambda" (customers/sec).
    variable ArrRate

    # Mean service rate or "mu" (customers/sec)
    variable ServRate

    #
    # Simulation variables.
    #

    # Mean interval between arrivals.
    variable IntMean

    # Mean service duration.
    variable ServMean

    # Simulated time.
    variable Time

    # Time interval.
    variable TimeDelta

    # Next interval.
    variable Interval

    # Time of next arrival.
    variable NextArr

    # Next service duration.
    variable Service

    # Time of next end of service.
    variable NextEos

    # Service busy.
    variable ServBusy

    # Queue of arrival times.
    variable Queue

    # Queue length.
    variable Qlen

    # Arrival time of the served customer.
    variable ArrServed

    # Waiting time.
    variable WaitTime

    # Time in system.
    variable SystemTime

    # Identifier of the next delayed run command to execute.
    variable RunId

    #
    # Statistic variables.
    #

    # Number of arrivals.
    variable NumArr

    # Number of not waiting arrivals.
    variable NoWait

    # Minimum arrival interval.
    variable MinInt

    # Maximum arrival interval.
    variable MaxInt

    # Number of completed services.
    variable NumServ

    # Minimum service duration.
    variable MinServ

    # Maximum service duration.
    variable MaxServ

    # Maximum queue length.
    variable MaxQlen

    # Maximum waiting time.
    variable MaxWait

    # Maximum time in system.
    variable MaxSystem

    #
    # Intermediate statistic variables.
    #

    # Queue length integral.
    variable QlenSum

    # Service busy integral.
    variable ServBusySum

    # Sum of waiting times.
    variable WaitTimeSum

    # Sum of times in system.
    variable SystemTimeSum

    #
    # Average values and probabilities.
    #

    # Server utilization.
    variable RoSim

    # Idle probability.
    variable P0Sim

    # Customers in system.
    variable LsSim

    # Customers in queue.
    variable LqSim

    # Time in system.
    variable WsSim

    # Time in queue.
    variable WqSim

    #
    # Limit average values and probabilities.
    #

    # Server utilization.
    variable RoLimit

    # Idle probability.
    variable P0Limit

    # Customers in system.
    variable LsLimit

    # Customers in queue.
    variable LqLimit

    # Time in system.
    variable WsLimit

    # Time in queue.
    variable WqLimit
}

# mm1Queue::expRand --
#
#        Generate a pseudo-random floating-point with an exponential
#        distribution function.
#
# Arguments:
#        mean       mean value or inverse of exponential distribution
#                   parameter, a floating-point number.
# Results:
#        Returns a pseudo-random floating-point number.

proc mm1Queue::expRand {mean} {
    return [expr {-$mean*log(rand())}]
}

# mm1Queue::mm1set --
#
#       Provides a controlled interface (alike the set(n) global command) for
#       reading and writing simulation variables.
#
# Arguments:
#       varName     name of the variable to read or write. All variables
#                   belonging to the mm1Queue namespace can be read, but only
#                   arrRate and servRate can be setted and their values are
#                   checked for correctness. Note that all variables name should
#                   start with a lower-case letter.
#       args        (optional). The new value to set variable varName to.
# Results:
#       Returns the value of variable varName. If value is specified, then set
#       the value of varName to args and return its value.

proc mm1Queue::mm1set {varName args} {
    set varName [string toupper $varName 0 0]
    switch [llength $args] {
        0 {
            # Read variable.

            variable $varName

            return [set $varName]
        }

        1 {
            # Write variable.

            switch -- $varName {
                ArrRate {
                    # Checks the current average number of arrivals per time unit.
                    if {![string is double $args] || $args <= 0} {
                        error "arrival rate is not a positive value"
                    }

                    variable ArrRate $args
                    variable IntMean [expr {1/double($ArrRate)}]

                    catch LimitValues

                    return $ArrRate
                }

                ServRate {
                    # Checks the current average number of completed services per time unit.
                    if {![string is double $args] || $args <= 0} {
                        error "service rate is not a positive value"
                    }

                    variable ServRate $args
                    variable ServMean [expr {1/double($ServRate)}]

                    catch LimitValues

                    return $ServRate
                }

                default {
                    error "can't set variable \"$varName\""
                }
            }
        }

        default {
            error {wrong # args: should be "mm1set varName ?newValue?"}
        }
    }
}

# mm1Queue::LimitValues --
#
#       Computes limit average values and probabilities of the M/M/1 theoretical model.
#
# Results:
#       Returns an empty string.

proc mm1Queue::LimitValues {} {
    if {[isStable]} {
        variable ArrRate
        variable ServRate
        variable RoLimit [expr {double($ArrRate)/$ServRate}]
        variable P0Limit [expr {1-$RoLimit}]
        variable LsLimit [expr {$RoLimit/$P0Limit}]
        variable LqLimit [expr {$LsLimit*$RoLimit}]
        variable WsLimit [expr {1/double($ServRate-$ArrRate)}]
        variable WqLimit [expr {$WsLimit*$RoLimit}]
    } else {
        variable RoLimit {}
        variable P0Limit {}
        variable LsLimit {}
        variable LqLimit {}
        variable WsLimit {}
        variable WqLimit {}
    }

    return
}

# mm1Queue::isStable --
#
#       Test an mm1 system stability. Should be done before running the simulation
#       and after every change of the system parameters to avoid to use up
#       memory or otherwise hang the system.
#
# Arguments:
#       state   (optional) a list containing the parameters of the system to
#               test of the form {ArrRate ServRate}. If empty, the parameters of the
#               current system are used.
# Results:
#       Returns the boolean 1 if the system is stable, 0 otherwise.

proc mm1Queue::isStable {{state ""}} {
    if {$state != ""} {
        set ArrRate [lindex $state 0]
        set ServRate [lindex $state 1]
    } else {
        variable ArrRate
        variable ServRate
    }
    return [expr {$ArrRate<$ServRate}]
}

# mm1Queue::reset --
#
#       Initialize a new simulation.
#
# Results:
#       Returns an empty string.

proc mm1Queue::reset {} {
    variable IntMean
    variable Time 0
    variable PrevTime 0
    variable NextArr
    variable NextEos
    variable ServBusy 0
    variable Queue {}
    variable Qlen 0
    variable RunId {}
    variable NumArr 0
    variable NoWait 0
    variable MinInt {}
    variable MaxInt 0
    variable NumServ 0
    variable MinServ {}
    variable MaxServ 0
    variable MaxQlen 0
    variable MaxWait 0
    variable MaxSystem 0
    variable QlenSum 0
    variable ServBusySum 0
    variable WaitTimeSum 0
    variable SystemTimeSum 0

    # Scheduling first arrival.
    set NextArr [expRand $IntMean]

    # This means no service scheduled.
    set NextEos [expr {$NextArr+1}]
}

# mm1Queue::run --
#
#       Run a simulation continuosly in the background. Also resume a
#       simulation that has been paused from exactly where it was left.
#
# Results:
#        Returns an empty string.

proc mm1Queue::run {} {
    variable RunId

    oneStep
    set RunId [after idle [list after 0 [namespace code run]]]
    return
}

# mm1Queue::pause --
#
#       Pause a simulation.
#
# Results:
#       Returns an empty string.

proc mm1Queue::pause {} {
    variable RunId

    after cancel $RunId
}

# mm1Queue::oneStep --
#
#       Do one step of a simulation and return.
#
# Results:
#       Returns an empty string.

proc mm1Queue::oneStep {} {
    variable IntMean
    variable ServMean
    variable Time
    variable PrevTime
    variable Interval
    variable NextArr
    variable Service
    variable NextEos
    variable ServBusy
    variable Queue
    variable Qlen
    variable ArrServed
    variable WaitTime
    variable SystemTime

    if {$NextArr < $NextEos} {
        # Arrival to occur.

        # Advance time.
        set PrevTime $Time
        set Time $NextArr

        # Update statistics.
        CumulativeStats

        # Scheduling the next arrival.
        set Interval [expRand $IntMean]
        set NextArr [expr {$NextArr+$Interval}]

        # Update statistics.
        ArrStats

        if {$ServBusy} {
            # Queuing necessary.

            # Record arrival time in queue.
            lappend Queue $Time
            incr Qlen

            # Update statistics.
            QueueStats
        } else {
            # Service not busy.

            set ServBusy 1

            # Scheduling the end of service.
            set Service  [expRand $ServMean]
            set NextEos [expr {$Time+$Service}]

            set ArrServed $Time

            # Update statistics.
            NoWaitStats
        }
    } else {
        # End of service to occur.

        # Advance time.
        set PrevTime $Time
        set Time $NextEos

        # Update statistics.
        CumulativeStats

        set ServBusy 0

        set SystemTime [expr {$Time-$ArrServed}]

        # This means no service scheduled.
        set NextEos [expr {$NextArr + 1}]

        # Update statistics.
        ServStats

        if {$Qlen > 0} {
            # Start the next service.

            set ServBusy 1

            # Scheduling the end of service.
            set Service  [expRand $ServMean]
            set NextEos [expr {$Time+$Service}]

            set ArrServed [lindex $Queue 0]

            set WaitTime [expr {$Time-$ArrServed}]

            set Queue [lreplace $Queue 0 0]
            incr Qlen -1

            # Update statistics.
            NextStats
        }
    }

    return
}

# mm1Queue::CumulativeStats --
#
#       Update statistic integrals at every step.
#
# Results:
#       Returns an empty string.

proc mm1Queue::CumulativeStats {} {
    variable Time
    variable PrevTime
    variable Qlen
    variable QlenSum
    variable ServBusy
    variable ServBusySum

    set delta [expr {$Time-$PrevTime}]
    set QlenSum [expr {$QlenSum+$Qlen*$delta}]
    set ServBusySum [expr {$ServBusySum+$ServBusy*$delta}]

    return
}

# mm1Queue::ArrStats --
#
#       Update statistics when a new arrival occurs.
#
# Results:
#       Returns an empty string.

proc mm1Queue::ArrStats {} {
    variable NumArr
    variable Interval
    variable MinInt
    variable MaxInt

    # Count this arrival.
    incr NumArr

    # Update minimum and maximum arrival interval.
    if {$MinInt == "" || $Interval<$MinInt} {
        set MinInt $Interval
    }
    if {$Interval>$MaxInt} {
        set MaxInt $Interval
    }

    return
}

# mm1Queue::NoWaitStats --
#
#       Update statistics when a new arrival occurs and the service is not
#       busy.
#
# Results:
#       Returns an empty string.

proc mm1Queue::NoWaitStats {} {
    variable NoWait
    variable Service
    variable MinServ
    variable MaxServ

    # Count this not waiting arrival.
    incr NoWait

    # Update minimum and maximum service time.
    if {$MinServ == "" || $Service<$MinServ} {
        set MinServ $Service
    }
    if {$Service>$MaxServ} {
        set MaxServ $Service
    }

    return
}

# mm1Queue::ServStats --
#
#       Update statistics when an end of service occurs.
#
# Results:
#       Returns an empty string.

proc mm1Queue::ServStats {} {
    variable NumServ
    variable SystemTime
    variable SystemTimeSum
    variable MaxSystem

    # Count this service.
    incr NumServ

    set SystemTimeSum [expr {$SystemTimeSum+$SystemTime}]

    # Update maximum system time.
    if {$SystemTime>$MaxSystem} {
        set MaxSystem $SystemTime
    }

    return
}

# mm1Queue::QueueStats --
#
#       Update statistics when the next service is started.
#
# Results:
#       Returns an empty string.

proc mm1Queue::QueueStats {} {
    variable Qlen
    variable MaxQlen

    # Update maximum queue length.
    if {$Qlen>$MaxQlen} {
        set MaxQlen $Qlen
    }

    return
}

# mm1Queue::NextStats --
#
#       Update statistics when an end of service occurs and there is another
#       customer in the queue to serve.
#
# Results:
#       Returns an empty string.

proc mm1Queue::NextStats {} {
    variable WaitTime
    variable WaitTimeSum
    variable MaxWait

    set WaitTimeSum [expr {$WaitTimeSum+$WaitTime}]

    # Update maximum waiting time.
    if {$WaitTime>$MaxWait} {
        set MaxWait $WaitTime
    }

    return
}

# mm1Queue::evaluate --
#
#       Evaluate simulation results using intermediate and statistic variables.
#
# Results:
#       Returns an empty string.

proc mm1Queue::evaluate {} {
    variable Time
    variable ServBusy
    variable NumArr
    variable NumServ
    variable NoWait
    variable QlenSum
    variable ServBusySum
    variable WaitTimeSum
    variable SystemTimeSum

    variable RoSim [expr {$ServBusySum/$Time}]
    variable P0Sim [expr {double($NoWait)/$NumArr}]
    variable LqSim [expr {$QlenSum/$Time}]
    variable LsSim [expr {$LqSim+$RoSim}]
    set n $NumServ
    if {$ServBusy} {
        # Count also the currently served customer.
        incr n
    }
    variable WqSim [expr {$WaitTimeSum/$n}]
    if {$NumServ>0} {
        variable WsSim [expr {$SystemTimeSum/$NumServ}]
    } else {
        variable WsSim {}
    }

    return
}

namespace eval mm1Queue {
    # Set default system parameters.
    mm1set arrRate 0.05
    mm1set servRate 0.10

    # Initial reset.
    reset
}

# End of file
# vim: set ts=4 sw=4 noet:
